package edu.fu_tdg.lodib.mapping;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.hp.hpl.jena.query.Dataset;

import edu.fu_tdg.lodib.mapping.sparql.factory.MappingFactory;
import edu.fu_tdg.lodib.mapping.sparql.factory.MappingType;
import edu.fu_tdg.lodib.mapping.sparql.factory.TechnologyType;
import edu.fu_tdg.lodib.mapping.xml.AddLangTag;
import edu.fu_tdg.lodib.mapping.xml.Aggregation;
import edu.fu_tdg.lodib.mapping.xml.ChangeDatatype;
import edu.fu_tdg.lodib.mapping.xml.Deresourcify;
import edu.fu_tdg.lodib.mapping.xml.Mapping;
import edu.fu_tdg.lodib.mapping.xml.NToOne;
import edu.fu_tdg.lodib.mapping.xml.Namespace;
import edu.fu_tdg.lodib.mapping.xml.OneToOne;
import edu.fu_tdg.lodib.mapping.xml.RemoveLangTag;
import edu.fu_tdg.lodib.mapping.xml.RenameClass;
import edu.fu_tdg.lodib.mapping.xml.RenameClassBasedOnProperty;
import edu.fu_tdg.lodib.mapping.xml.RenameClassBasedOnValue;
import edu.fu_tdg.lodib.mapping.xml.RenameProperty;
import edu.fu_tdg.lodib.mapping.xml.Resourcify;
import edu.fu_tdg.lodib.mapping.xml.ReverseProperty;
import edu.fu_tdg.lodib.mapping.xml.URIToValue;
import edu.fu_tdg.lodib.mapping.xml.ValueToURI;
import edu.fu_tdg.lodib.utils.Utils;
import edu.fu_tdg.lodib.validator.DataLoader;

public class DataTranslator {
	private Map<String, String> namespaces = new HashMap<String, String>();
	private List<DataTranslationMapping> mappings = new ArrayList<DataTranslationMapping>();

    public void performDataTranslation(File xml, File sourceRDF, File targetRDF, TechnologyType tech) throws DataTranslationException, IOException {
        File tempFile = File.createTempFile("lodib-", "tmp");
        tempFile.deleteOnExit();
        performDataTranslation(xml, sourceRDF, tempFile, targetRDF, tech);
    }
	
	public void performDataTranslation(File xml, File sourceRDF, File tempFile, File targetRDF, TechnologyType tech) throws DataTranslationException {
		// Transform XML file into internal objects
		try {
            targetRDF.delete();
			Mapping xmlObject = Utils.unmarshall(Mapping.class, new FileInputStream(xml));
			
			// Add namespaces to map
			for (Namespace nm : xmlObject.getNamespace())
				namespaces.put(nm.getPrefix(), nm.getValue());
			
			// Parse mappings
			for (RenameClass rcl : xmlObject.getRenameClass())
				mappings.add(parseRenameClass(rcl, tech));
			for (RenameProperty rcp : xmlObject.getRenameProperty())
				mappings.add(parseRenameProperty(rcp, tech));
			for (RenameClassBasedOnProperty rcp : xmlObject.getRenameClassBasedOnProperty())
				mappings.add(parseRenameClassBasedOnProperty(rcp, tech));
			for (RenameClassBasedOnValue rcv : xmlObject.getRenameClassBasedOnValue())
				mappings.add(parseRenameClassBasedOnValue(rcv, tech));
			for (ReverseProperty rvp : xmlObject.getReverseProperty())
				mappings.add(parseReverseProperty(rvp, tech));
			for (Resourcify rsc : xmlObject.getResourcify())
				mappings.add(parseResourcify(rsc, tech));
			for (Deresourcify rsc : xmlObject.getDeresourcify())
				mappings.add(parseDeresourcify(rsc, tech));
			for (OneToOne oto : xmlObject.getOneToOne())
				mappings.add(parseOneToOne(oto, tech));
			for (URIToValue utv : xmlObject.getURIToValue())
				mappings.add(parseURIToValue(utv, tech));
			for (ValueToURI vtu : xmlObject.getValueToURI())
				mappings.add(parseValueToURI(vtu, tech));
			for (ChangeDatatype cd : xmlObject.getChangeDatatype())
				mappings.add(parseChangeDatatype(cd, tech));
			for (AddLangTag alt : xmlObject.getAddLangTag())
				mappings.add(parseAddLangTag(alt, tech));
			for (RemoveLangTag rlt : xmlObject.getRemoveLangTag())
				mappings.add(parseRemoveLangTag(rlt, tech));
			for (NToOne nto : xmlObject.getNToOne())
				mappings.add(parseNToOne(nto, tech));
			for (Aggregation agg : xmlObject.getAggregation())
				mappings.add(parseAggregation(agg, tech));

            Dataset sourceDataset = DataLoader.loadDataset(sourceRDF.toURI().toString());
			// Execute mappings
			targetRDF.createNewFile();
			for (DataTranslationMapping m : mappings) {
                System.out.println("Executing mapping: " + m.getClass().getSimpleName());
				m.execute(sourceDataset, tempFile);
				Utils.copy(tempFile, targetRDF, true);
			}
		} catch (Exception oops) {
			throw new DataTranslationException("Error when performing data translation: " + oops.getMessage(), oops);
		}
	}
	
	private DataTranslationMapping parseRenameClass(RenameClass rcl, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceClass(rcl.getSourceClass());
		input.setTargetClass(rcl.getTargetClass());
		
		return MappingFactory.createMapping(MappingType.RENAME_CLASS, tech, input);
	}
	
	private DataTranslationMapping parseRenameProperty(RenameProperty rcp, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(rcp.getSourceProperty());
		input.setTargetProperty(rcp.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.RENAME_PROPERTY, tech, input);
	}
	
	private DataTranslationMapping parseRenameClassBasedOnProperty(RenameClassBasedOnProperty rcp, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceClass(rcp.getSourceClass());
		input.setSourceProperty(rcp.getSourceProperty());
		input.setTargetClass(rcp.getTargetClass());
		
		return MappingFactory.createMapping(MappingType.RENAME_CLASS_BASED_ON_PROPERTY, tech, input);
	}
	
	private DataTranslationMapping parseRenameClassBasedOnValue(RenameClassBasedOnValue rcv, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceClass(rcv.getSourceClass());
		input.setSourceProperty(rcv.getSourceProperty());
		input.setValue(rcv.getSourceValue());
		input.setSourceType(rcv.getSourceType());
		input.setTargetClass(rcv.getTargetClass());
		
		return MappingFactory.createMapping(MappingType.RENAME_CLASS_BASED_ON_VALUE, tech, input);
	}
	
	private DataTranslationMapping parseReverseProperty(ReverseProperty rvp, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(rvp.getSourceProperty());
		input.setTargetProperty(rvp.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.REVERSE_PROPERTY, tech, input);
	}
	
	private DataTranslationMapping parseResourcify(Resourcify rsc, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(rsc.getSourceProperty());
		input.setTargetProperty1(rsc.getTargetProperty1());
		input.setTargetProperty2(rsc.getTargetProperty2());
		
		return MappingFactory.createMapping(MappingType.RESOURCIFY, tech, input);
	}
	
	private DataTranslationMapping parseDeresourcify(Deresourcify rsc, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty1(rsc.getSourceProperty1());
		input.setSourceProperty2(rsc.getSourceProperty2());
		input.setTargetProperty(rsc.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.DERESOURCIFY, tech, input);
	}
	
	private DataTranslationMapping parseOneToOne(OneToOne oto, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(oto.getSourceProperty());
		input.setTargetProperty(oto.getTargetProperty());
		input.setFunction(oto.getFunction());
		
		return MappingFactory.createMapping(MappingType.ONE_TO_ONE, tech, input);
	}
	
	private DataTranslationMapping parseURIToValue(URIToValue utv, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(utv.getSourceProperty());
		input.setTargetProperty(utv.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.URI_TO_VALUE, tech, input);
	}
	
	private DataTranslationMapping parseValueToURI(ValueToURI vtu, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(vtu.getSourceProperty());
		input.setTargetProperty(vtu.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.VALUE_TO_URI, tech, input);
	}
	
	private DataTranslationMapping parseChangeDatatype(ChangeDatatype cd, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(cd.getSourceProperty());
		if (cd.getSourceType() != null)
			input.setSourceType(cd.getSourceType());
		input.setTargetProperty(cd.getTargetProperty());
		if (cd.getTargetType() != null)
			input.setTargetType(cd.getTargetType());
		
		return MappingFactory.createMapping(MappingType.CHANGE_DATATYPE, tech, input);
	}
	
	private DataTranslationMapping parseAddLangTag(AddLangTag alt, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(alt.getSourceProperty());
		input.setLangTag(alt.getLanguageTag());
		input.setTargetProperty(alt.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.ADD_LANG_TAG, tech, input);
	}
	
	private DataTranslationMapping parseRemoveLangTag(RemoveLangTag rlt, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceProperty(rlt.getSourceProperty());
		input.setLangTag(rlt.getLanguageTag());
		input.setTargetProperty(rlt.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.REMOVE_LANG_TAG, tech, input);
	}
	
	private DataTranslationMapping parseNToOne(NToOne nto, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		List<String> sourceProperties = new ArrayList<String>();
		sourceProperties.add(nto.getSourceProperty1());
		sourceProperties.add(nto.getSourceProperty2());
		
		input.setNSourceProperties(sourceProperties);
		input.setTargetProperty(nto.getTargetProperty());
		input.setFunction(nto.getFunction());
		
		return MappingFactory.createMapping(MappingType.N_TO_ONE, tech, input);
	}
	
	private DataTranslationMapping parseAggregation(Aggregation agg, TechnologyType tech) {
		InputDataTranslationMapping input = new InputDataTranslationMapping(namespaces);
		
		input.setSourceClass(agg.getSourceClass());
		input.setSourceProperty(agg.getSourceProperty());
		input.setTargetClass(agg.getTargetClass());
		input.setTargetProperty(agg.getTargetProperty());
		
		return MappingFactory.createMapping(MappingType.AGGREGATION, tech, input);
	}
	
}
