package edu.fu_tdg.lodib.validator

import java.io.{FileWriter, BufferedWriter}
import collection.mutable.ArrayBuffer
import com.hp.hpl.jena.query.{ResultSet, QuerySolution, QueryExecutionFactory}

/**
 * Created by IntelliJ IDEA.
 * User: andreas
 * Date: 1/19/12
 * Time: 6:51 PM
 * To change this template use File | Settings | File Templates.
 */

class Validator(resultFile: String, correctDataset: String, datasetToValidate: String, verbose: Boolean = true) {
  private var counter = 0
  private val log = new BufferedWriter(new FileWriter(resultFile))
  val correct = DataLoader.loadDataset(correctDataset)
  val toValidate = DataLoader.loadDataset(datasetToValidate)
  var lastMissing = ""
  var lastAdditional = ""

  def validate() {
    println("Validating...")
    var problematicMappingsCounter = 0
    for((query, varNames) <- queries.values)
      if(!validateByQuery(query, varNames))
        problematicMappingsCounter += 1
    if(problematicMappingsCounter==0)
      println("Validation successful, no problems found!")
    else {
      val mappings = if(problematicMappingsCounter==1) "mapping" else "mappings"
      println("The validation discovered problems, " + problematicMappingsCounter + " questionable/missing " + mappings + " uncovered!")
    }
    log.flush()
    log.close()
  }

  private def validateByQuery(query: String, varNames: Seq[String]): Boolean = {
    val correctResults = QueryExecutionFactory.create(query, correct).execSelect()
    val toValidateResults = QueryExecutionFactory.create(query, toValidate).execSelect()
    var oneFull = correctResults.hasNext || toValidateResults.hasNext
    var correctSolution: QuerySolution = null
    var toValidateSolution: QuerySolution = null
    val verboseProblems = new StringBuilder
    counter = 0

    // Special cases

    if(!correctResults.hasNext) {
      if(verbose)
        log.append("No data found in validation data set for validation query: " + query + "\n")
      return true
    }

    if(!toValidateResults.hasNext) {
      log.append("Problem: No mapped data found in test data set for the following validation query: " + query + "\n")
      return false
    }

    while(oneFull) {
      if((correctSolution==null&&toValidateSolution==null) || equal(correctSolution, toValidateSolution, varNames)) {
        correctSolution = updateCorrectSolution(correctResults)
        toValidateSolution = updateToValidateSolution(toValidateResults)
      } else {
        if(correctSolution==null)
          toValidateSolution = updateToValidateSolution(toValidateResults)
        else if(toValidateSolution==null)
          correctSolution = updateCorrectSolution(correctResults)
        else if(smaller(correctSolution, toValidateSolution, varNames))
          correctSolution = updateCorrectSolution(correctResults)
        else
          toValidateSolution = updateToValidateSolution(toValidateResults)
      }

      if(smaller(correctSolution, toValidateSolution, varNames))
        logMissingResult(correctSolution, verboseProblems)

      if(smaller(toValidateSolution, correctSolution, varNames))
          logAdditionalResult(toValidateSolution, verboseProblems)
      oneFull = correctResults.hasNext || toValidateResults.hasNext
    }
    if(correctSolution!=null&&toValidateSolution!=null) {
      if(smaller(correctSolution, toValidateSolution, varNames))
        logAdditionalResult(toValidateSolution, verboseProblems)
      else if(smaller(toValidateSolution, correctSolution, varNames))
        logMissingResult(correctSolution, verboseProblems)
    }
    if(counter==0)
      return true
    else {
      val problems = if(counter==1) "problem" else "problems"
      log.append("Problem: Found " + counter + " " + problems + " in test data set for validation query: " + query + "\n")
      if(verbose)
        log.append(verboseProblems.toString())
      return false
    }
  }

  private def logMissingResult(solution: QuerySolution, verboseProblems: StringBuilder) {
    val error = "  Missing result detected! Values: " + solution.toString + "\n"
    if(error!=lastMissing) {
      if(verbose)
        verboseProblems.append(error)
      counter += 1
    }
    lastMissing = error
  }

  private def logAdditionalResult(solution: QuerySolution, verboseProblems: StringBuilder) {
    val error = "  Unexpected result found! Values: " + solution.toString + "\n"
    if(error!=lastAdditional) {
      if(verbose)
        verboseProblems.append(error)
      counter += 1
    }
    lastAdditional = error
  }

  private def updateCorrectSolution(correctResults: ResultSet): QuerySolution = {
    if (correctResults.hasNext)
      return correctResults.next()
    return null
  }

  private def updateToValidateSolution(toValidateResults: ResultSet): QuerySolution = {
    if (toValidateResults.hasNext)
      return toValidateResults.next()
    return null
  }

  /**
   * Compares each position of the query solution if the value is smaller for the first query solution
   */
  private def smaller(qSol1: QuerySolution, qSol2: QuerySolution, varNames: Seq[String]): Boolean = {
    if(qSol2==null)
      return true
    if(qSol1==null)
      return false
    for(varName <- varNames) {
      val var1 = qSol1.get(varName).toString
      val var2 = qSol2.get(varName).toString
      if(var1 < var2)
        return true
      else if(var2 > var1)
        return false
    }
    return false
  }

  private def equal(qSol1: QuerySolution, qSol2: QuerySolution, varNames: Seq[String]): Boolean = {
    if(qSol1==null||qSol2==null)
      return false
    for(varName <- varNames) {
      val var1 = qSol1.get(varName)
      val var2 = qSol2.get(varName)
      if(!((var1.isLiteral && var2.isLiteral) || (var1.isURIResource && var2.isURIResource)))
        return false
      if(var1.isURIResource && var1.asResource().getURI!=var2.asResource().getURI)
        return false
      else if(var1.isLiteral) {
        val lit1 = var1.asLiteral()
        val lit2 = var2.asLiteral()
        if(lit1.getDatatype!=lit2.getDatatype)
          return false
        if(lit1.getLanguage!=lit2.getLanguage)
          return false
        if(lit1.getLexicalForm!=lit2.getLexicalForm)
          return false
      }
    }
    return true
  }

  private def getVarNames(qs: QuerySolution): Seq[String] = {
    val vnames = qs.varNames()
    val result = new ArrayBuffer[String]()
    while(vnames.hasNext)
      result.append(vnames.next())
    return result
  }
}

object Validator {
  def validate(correctDataset: String, datasetToValidate: String, resultFile: String, verbose: Boolean=true) {
    val validator = new Validator(resultFile, correctDataset, datasetToValidate, verbose)
    validator.validate()
  }

  def main(args: Array[String]) {
    if(args.length < 3 || args(0)=="-v" && args.length < 4) {
      println("Parameters: [-v] <correct data set> <test data set> <output file>")
      println("  -v: optional, verbose output")
      sys.exit(1)
    }
    if(args(0)=="-v")
      Validator.validate(args(1),args(2),args(3))
    else
      Validator.validate(args(0),args(1),args(2), false)
  }
}

object queries {

  /*
   * Maybe add filters for: lang, literal/data type, IRI etc.
   * FILTER langMatches(lang(?label), "en")
   * Other functions: datatype, isIRI, isLiteral
   *
   */
  val values = List[Pair[String, Seq[String]]](
    // check classes
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Product> } order by ?s", List("s")),
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Review> } order by ?s", List("s")),
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/OutdatedProduct> } order by ?s", List("s")),
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Person> } order by ?s", List("s")),
// Birth as a helper resource doesn't need to be checked. This is done with tgt:birthDate
//    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Birth> } order by ?s", List("s")),
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Reviewer> } order by ?s", List("s")),
    ("select ?s { ?s a <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/Review> } order by ?s", List("s")),
    // check properties
    ("select ?s ?o { ?s <http://www.w3.org/2000/01/rdf-schema#label> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>) } order by ?s ?o ", List("s","o")),
    ("select ?s ?o { ?s <http://www.w3.org/2000/01/rdf-schema#comment> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/productVendor> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/productPrice> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#double>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/productHomepage> ?o FILTER(isIRI(?o))} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/totalReviews> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#integer>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/productReview> ?o } order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/text> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://purl.org/dc/elements/1.1/date> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#date>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/reviewSubject> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/name> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://xmlns.com/foaf/0.1/mbox> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/bio> ?o FILTER langMatches(lang(?o), 'en')} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/personHomepage> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#string>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/initialDate> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#date>)} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/author> ?o} order by ?s ?o", List("s","o")),
    ("select ?s ?o { ?s <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/birth> ?b . ?b <http://www4.wiwiss.fu-berlin.de/bizer/lodib/vocabulary/target/birthDate> ?o FILTER(datatype(?o)=<http://www.w3.org/2001/XMLSchema#date>)} order by ?s ?o", List("s","o"))
  )
}